#!/usr/bin/env python
"""
The radio abstraction for COS 463
"""

import logging

import numpy
import SoapySDR

LOG = logging.getLogger(__name__)

class SchmidlCox():
    """ Implements the Schmidl-Cox preamble detection algorithm. """

    def __init__(self, short_preamble_len, threshold=0.5):
        self.p = 0
        self.r = 0
        self.threshold = threshold
        self.short_preamble_len = short_preamble_len
        self.old_buffer = numpy.zeros(2 * short_preamble_len, dtype=numpy.complex)

    def process_buffer(self, buff):
        """
        Returns either None or an index indicating the start of the preamble.
        The index may be negative.
        """
        signal = numpy.concatenate((self.old_buffer, buff))
        self.old_buffer = signal[-2 * self.short_preamble_len:]

        p_increments = signal[self.short_preamble_len:-self.short_preamble_len] * \
            signal[2 * self.short_preamble_len:].conj() - signal[:-2 * self.short_preamble_len] * \
            signal[self.short_preamble_len:-self.short_preamble_len].conj()
        Ps = numpy.cumsum(p_increments) + self.p
        self.p = Ps[-1]

        r_increments = (abs(signal[2 * self.short_preamble_len:]) ** 2) -\
            (abs(signal[:-2 * self.short_preamble_len]) ** 2)
        Rs = 0.5 * numpy.cumsum(r_increments) + self.r
        self.r = Rs[-1]

        Ms = (abs(Ps)**2) / (Rs**2)

        max_idx = numpy.argmax(Ms)
        LOG.debug("Max value of M: %s", Ms[max_idx])
        if Ms[max_idx] > self.threshold:
            signal_idx = max_idx - len(self.old_buffer) + 1
            return signal_idx


class Radio():

    def __init__(self, *args, **kwargs):
        self.sdr = SoapySDR.Device(*args, **kwargs)

    def set_sample_rate(self, sample_rate_hz):
        self.bandwidth_hz = sample_rate_hz
        self.sdr.setSampleRate(SoapySDR.SOAPY_SDR_RX, 0, sample_rate_hz)
        self.sdr.setSampleRate(SoapySDR.SOAPY_SDR_TX, 0, sample_rate_hz)

    def set_center_frequency(self, freq_hz):
        self.sdr.setFrequency(SoapySDR.SOAPY_SDR_RX, 0, freq_hz)
        self.sdr.setFrequency(SoapySDR.SOAPY_SDR_TX, 0, freq_hz)

    def _set_gain(self, gain, direction):
        max_gain = sum(self.sdr.getGainRange(direction, 0, gain_type).maximum()
            for gain_type in self.sdr.listGains(direction, 0))
        self.sdr.setGain(direction, 0, gain * max_gain)

    def set_receive_gain(self, gain):
        """
        `gain` is a value between 0 and 1. Note that `gain` corresponds to a value in decibels,
        so the actual gain is exponential in the value of `gain`.
        """
        self._set_gain(gain, SoapySDR.SOAPY_SDR_RX)

    def set_transmit_gain(self, gain):
        """
        `gain` is a value between 0 and 1. Note that `gain` corresponds to a value in decibels,
        so the actual gain is exponential in the value of `gain`.
        """
        self._set_gain(gain, SoapySDR.SOAPY_SDR_TX)

    def receive_frame(self, short_preamble_len, message_len, threshold=0.5):
        schmidl_cox = SchmidlCox(short_preamble_len, threshold)

        rx_stream = self.sdr.setupStream(SoapySDR.SOAPY_SDR_RX, "CF32", [0])
        mtu = self.sdr.getStreamMTU(rx_stream)
        self.sdr.activateStream(rx_stream)

        try:
            chunk = numpy.zeros(mtu, dtype=numpy.complex64)
            while True:
                rx_resp = self.sdr.readStream(
                    rx_stream, [chunk], numElems=len(chunk), timeoutUs=int(1e6 * mtu / self.bandwidth_hz))
                if rx_resp.ret != len(chunk):
                    raise RuntimeError(f"Receive failed: {SoapySDR.errToStr(rx_resp.ret)}")

                dc_removed = chunk - numpy.mean(chunk)
                start_idx = schmidl_cox.process_buffer(dc_removed)
                if start_idx is not None:
                    # TODO handle negative offsets or large message lengths
                    received_frame = dc_removed[start_idx:start_idx + short_preamble_len * 2 + message_len]
                    return received_frame

        finally:
            self.sdr.deactivateStream(rx_stream)
            self.sdr.closeStream(rx_stream)

    def transmit_signal(self, signal):
        """
        TODO: handle signals longer than the MTU
        TODO: Why do I need to send so much garbage before the actual signal is transmitted?
        """
        tx_stream = self.sdr.setupStream(SoapySDR.SOAPY_SDR_TX, "CF32", [0])

        mtu = self.sdr.getStreamMTU(tx_stream)
        padded_signal = numpy.zeros(mtu, dtype=numpy.complex64)
        padded_signal[-len(signal):] = signal

        # Need to repeat the signal this many times before it's actually transmitted!
        num_repeats = 4

        self.sdr.activateStream(tx_stream, numElems=num_repeats * len(padded_signal))

        for _ in range(num_repeats):
            tx_resp = self.sdr.writeStream(tx_stream, [padded_signal], numElems=len(padded_signal))
            if tx_resp.ret != len(padded_signal):
                raise RuntimeError("Transmit failed: {}".format(SoapySDR.errToStr(tx_resp.ret)))

        # Wait for transmission to finish
        self.sdr.readStreamStatus(
            tx_stream, timeoutUs=num_repeats * int(1e6 * len(padded_signal) / self.bandwidth_hz))

        self.sdr.deactivateStream(tx_stream)
        self.sdr.closeStream(tx_stream)
